# vim:sw=4:et
#############################################################################
# File : BackportsPolicyChecks.py
# Package : rpmlint
# Author : Scott Bahling
# Purpose : Policy checks for the openSUSE:Backports project
#############################################################################
from Filter import printError, printInfo, addDetails
import subprocess
import AbstractCheck
import Config
import shelve
import os
import re

bin_re = re.compile(r'^(?:/usr)?/s?bin/')
lib_re = re.compile(r'^(?:/usr)?/lib(?:64)?/(.+)')

DB_IMPORT_FAILED = 'BackportsDB_import_failed'

# Define our lint policies
policy = 'SUSE_Backports_policy-{}'.format
PKG_CONFLICT = policy('SLE_conflict')
NO_KMODS = policy('no_kernel_modules')
RM_PYTHON_CONFLICT = policy('remove_python_conflict')
NO_FORBIDDEN_FILES = policy('no_forbidden_files')

# Ignored policies
policy_ignore = 'SUSE_Backports_policy_ignored-{}'.format
IGNORE_ALT_CONFLICT = policy_ignore('SLE_alternatives_conflict')
IGNORE_FILE_CONFLICT = policy_ignore('SLE_files_conflict')
IGNORE_PKG_CONFLICT = policy_ignore('SLE_rpm_conflict')
IGNORE_SRC_CONFLICT = policy_ignore('SLE_src_conflict')

NONE_re = re.compile('NONE')


def _re_escape(string):
    """If string starts with '@' we interpret it as a regex definition
       after stripping the '@' leading character.
       If not, we interpret as a literal string by escaping any re special
       characters
    """
    if string[0] == '@':
        return string[1:]
    else:
        return re.escape(string)


IGNORE_CONFLICTS = Config.getOption('BackportsIgnoreConflicts', ())

IGNORE_FILES = {}
# pre-compile regex expresions
for pkgname, files in Config.getOption('BackportsIgnoreFileConflicts', {}).items():
    patterns = [_re_escape(i) for i in files]
    IGNORE_FILES[pkgname] = re.compile('(' + '|'.join(patterns) + ')')

patterns = [_re_escape(i) for i in Config.getOption('BackportsIgnoreAlternatives')]
if patterns:
    IGNORE_ALTERNATIVES = re.compile('(' + '|'.join(patterns) + ')')
else:
    IGNORE_ALTERNATIVES = None

patterns = [_re_escape(i) for i in Config.getOption('BackportsForbiddenFiles')]
if patterns:
    FORBIDDEN_FILES = re.compile('(' + '|'.join(patterns) + ')')
else:
    FORBIDDEN_FILES = None

mapfile = os.environ.get('BACKPORTS_DB_FILE',
                         '/usr/share/rpmlint/data/SLE-file-package-map.db')


class BackportsPolicyChecks(AbstractCheck.AbstractCheck):
    def __init__(self):
        AbstractCheck.AbstractCheck.__init__(self, 'BackportsPolicyChecks')

        try:
            d = shelve.open(mapfile, flag='r', protocol=2)
            self.filenames = d['filenames']
            self.pkgnames = d['pkgnames']
            d.close()
            self.db_imported = True
        except Exception:
            self.db_imported = False

    def _is_kmod(self, pkg, ko_file):
        try:
            printInfo(pkg, 'checking if %s is a kernel module' % ko_file)
            p = subprocess.Popen(['/sbin/modinfo', ko_file],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
            p.wait()
            modinfo, stderr = p.communicate()
            return True
        except Exception as e:
            printInfo(pkg, e)
            return False

    def _check_source(self, pkg):
        # we are only strict about binaries (FATE#323679)
        if pkg.name in self.pkgnames:
            printInfo(pkg,
                      IGNORE_SRC_CONFLICT,
                      'SRC package with name %s already in SLE' % pkg.name
                      )
            return

    def _ignore_file(self, pkgname, fname):
        any_pkg = IGNORE_FILES.get('ANY', NONE_re)
        this_pkg = IGNORE_FILES.get(pkgname, NONE_re)
        if any_pkg.match(fname) or this_pkg.match(fname):
            return True
        return False

    def _ignore_alternative(self, pkg, fname, pkgfile):
        if IGNORE_ALTERNATIVES is None:
            return False
        # the linkto path might be a relative path. We need to match against an
        # absolute path
        filepath = os.path.split(pkgfile.name)[0]
        abslinkto = os.path.abspath(os.path.join(filepath, pkgfile.linkto))
        if IGNORE_ALTERNATIVES.match(pkgfile.name) or IGNORE_ALTERNATIVES.match(abslinkto):
            return True
        return False

    def check(self, pkg):
        if not self.db_imported:
            printError(pkg, DB_IMPORT_FAILED, 'Failed to import %s' % mapfile)
            return

        if pkg.isSource():
            self._check_source(pkg)
            return

        if pkg.name in self.pkgnames and pkg.name.startswith('python'):
            os.remove(pkg.filename)
            printInfo(pkg, RM_PYTHON_CONFLICT, ' - Removing the package.')
            return

        if pkg.name in self.pkgnames:
            if pkg.name in IGNORE_CONFLICTS:
                printInfo(pkg, IGNORE_PKG_CONFLICT,
                          'Package with name %s is in SLE' % pkg.name
                          )
                return
            printError(pkg, PKG_CONFLICT,
                       'Package with name %s already in SLE' % pkg.name
                       )
            return

        conflicts = {}
        ignore_conflicts = {}
        ignore_alternatives = {}
        remove_python_conflicts = {}
        forbidden_files = []
        for fname, pkgfile in pkg.files().items():
            tocheck = [fname]
            if bin_re.match(fname):
                tocheck = [os.path.join(p, os.path.basename(fname))
                           for p in ('/bin', '/sbin', '/usr/bin', '/usr/sbin')]
            else:
                m = lib_re.match(fname)
                if m and not fname.startswith('/lib/modules'):
                    tocheck = [os.path.join(p, m.group(1))
                               for p in ('/lib', '/lib64', '/usr/lib', '/usr/lib64')]

            for fname in tocheck:

                if fname.endswith('.ko') and self._is_kmod(pkg, fname):
                    printError(pkg, NO_KMODS)
                    continue

                if FORBIDDEN_FILES and FORBIDDEN_FILES.match(fname):
                    if self._ignore_file(pkg.name, fname):
                        printInfo(pkg,
                                  'Whitelisted Forbidden File: %s' % fname
                                  )
                        continue
                    forbidden_files.append(fname)
                    continue

                conflict = self.filenames.get(fname, [])
                if not conflict:
                    continue

                if pkgfile.linkto:
                    indent = ' ' * (len(pkg.name + '.' + pkg.arch) + 24)
                    printInfo(pkg,
                              ('-- SLE Conflict -- File: %s\n%sLinkto: %s') %
                              (pkgfile.name, indent, pkgfile.linkto)
                              )
                else:
                    printInfo(pkg,
                              '-- SLE Conflict -- File: %s' % pkgfile.name
                              )

                if self._ignore_file(pkg.name, fname):
                    for package in conflict:
                        ignore_conflicts.setdefault(package, set()).add(fname)
                    continue

                if self._ignore_alternative(pkg, fname, pkgfile):
                    for package in conflict:
                        ignore_alternatives.setdefault(package, set()).add(fname)
                    continue

                # Handle python 2/3 conflicts when the package name has been
                # changed and not caught with the same package name check.
                if pkg.name.startswith('python'):
                    os.remove(pkg.filename)
                    printInfo(pkg, RM_PYTHON_CONFLICT, ' - Removing the package.')
                    return

                for package in conflict:
                    conflicts.setdefault(package, set()).add(fname)

        error = self.process_conflicts(ignore_alternatives)
        if error:
            printInfo(pkg, IGNORE_ALT_CONFLICT, error)

        error = self.process_conflicts(ignore_conflicts)
        if error:
            printInfo(pkg, IGNORE_FILE_CONFLICT, error)

        error = self.process_conflicts(conflicts)
        if error:
            printError(pkg, PKG_CONFLICT, error)

        error = self.process_conflicts(remove_python_conflicts)
        if error:
            printError(pkg, RM_PYTHON_CONFLICT, error)

        if forbidden_files:
            printError(pkg, NO_FORBIDDEN_FILES, '\n' + '\n'.join(forbidden_files))

    def process_conflicts(self, conflicts):
        if not conflicts:
            return ''

        msglines = []
        for package, files in conflicts.items():
            for fpath in sorted(files):
                msglines.append('%s: %s' % (package, fpath))

        return '\n' + '\n'.join(msglines)


check = BackportsPolicyChecks()

db_import_failed_msg = """BackportsPolicyCheck failed to import the package
list database. Something is broken with the rpmlint-backports checks. Please
open a report at bugzilla.suse.com under product 'openSUSE Backports'"""

package_conflict_msg = """As per Backports policy no package may have the same
name nor conflict with the file list of a package provided by SLE itself."""

ignore_package_conflict_msg = """The named package is part of SLE but was
whitelisted to allow override in the Backports project."""

ignore_src_conflict_msg = """The package has the same source package name as
a package in SLE. Source packages are allowed to conflict in the Backports project."""

ignore_files_conflict_msg = """The package contains files that conflict with
files from packages that are part of SLE but they are whitelisted to allow
override in the Backports project."""

ignore_alternatives_msg = """This package conflicts with files provided by SLE
via the 'alternatives' system and they are whitelisted to be allowed in the
Backports project."""

no_kernel_modules_msg = """The package contains kernel modules. By policy kernel
modules are prohibited in the openSUSE:Backports projects."""

rm_python_conflict_msg = """The package is a python package that conflicts with
a same package provided by SLE. Removing the package from the build set for
the case that we want to provide the "other half" of a python singlespec
build."""

forbidden_files_msg = """The package contains some files that are flagged
as forbidden in the Backports project."""

if Config.info:
    addDetails(PKG_CONFLICT, package_conflict_msg,
               NO_KMODS, no_kernel_modules_msg,
               RM_PYTHON_CONFLICT, rm_python_conflict_msg,
               NO_FORBIDDEN_FILES, forbidden_files_msg,
               DB_IMPORT_FAILED, db_import_failed_msg,
               IGNORE_PKG_CONFLICT, ignore_package_conflict_msg,
               IGNORE_FILE_CONFLICT, ignore_files_conflict_msg,
               IGNORE_ALT_CONFLICT, ignore_alternatives_msg,
               IGNORE_SRC_CONFLICT, ignore_src_conflict_msg,
               )
